//------------------------------------------------------------------------------
//
//	 ArchitectureCraft - Shape kinds
//
//------------------------------------------------------------------------------

package gcewing.architecture;

import java.util.*;
import static java.lang.Math.*;

import net.minecraft.block.*;
import net.minecraft.block.state.*;
import net.minecraft.entity.*;
import net.minecraft.entity.player.*;
import net.minecraft.init.*;
import net.minecraft.item.*;
import net.minecraft.nbt.*;
import net.minecraft.world.*;
import net.minecraft.tileentity.*;
import net.minecraft.util.*;

import static net.minecraft.util.EnumFacing.*;

import gcewing.architecture.BaseModClient.*;

//------------------------------------------------------------------------------

public abstract class ShapeKind {

	public Object[] profiles; // indexed by local face

	public abstract void renderShape(ShapeTE te,
		ITexture[] textures, IRenderTarget target, Trans3 t,
		boolean renderBase, boolean renderSecondary);
	
	public ItemStack newStack(Shape shape, Block materialBlock, int materialMeta, int stackSize) {
		ShapeTE te = new ShapeTE(shape, materialBlock, materialMeta);
		return BaseUtils.blockStackWithTileEntity(ArchitectureCraft.blockShape, stackSize, te);
	}

	public boolean orientOnPlacement(EntityPlayer player, ShapeTE te, ShapeTE nte, EnumFacing otherFace,
		Vector3 hit)
	{
		//boolean debug = !te.getWorld().isRemote;
		if (nte != null && !player.isSneaking()) {
			Object otherProfile = Profile.getProfileGlobal(nte.shape.kind, nte.side, nte.turn, otherFace);
			if (otherProfile != null) {
				EnumFacing thisFace = otherFace.getOpposite();
				for (int turn = 0; turn <= 4; turn++) {
					Object thisProfile = Profile.getProfileGlobal(te.shape.kind, nte.side, turn, thisFace);
					if (Profile.matches(thisProfile, otherProfile)) {
						//if (debug)
						//	System.out.printf("ShapeKind.orientOnPlacement: side %s turn %s\n", nte.side, turn);
						te.setSide(nte.side);
						te.setTurn(turn);
						return true;
					}
				}
			}
		}
		return false;
	}
	
	public boolean canPlaceUpsideDown() {
		return true;
	}

	public double sideZoneSize() {
		return 1/4d;
	}
	
	public boolean highlightZones() {
		return false;
	}
	
	public void onChiselUse(ShapeTE te, EntityPlayer player, EnumFacing face, Vector3 hit) {
		EnumFacing side = zoneHit(face, hit);
		//System.out.printf("ShapeKind.onChiselUse: face = %s, hit = %s, side = %s\n", face, hit, side);
		if (side != null)
			chiselUsedOnSide(te, player, side);
		else
			chiselUsedOnCentre(te, player);
	}
	
	public void chiselUsedOnSide(ShapeTE te, EntityPlayer player, EnumFacing side) {
		//System.out.printf("ShapeKind.chiselUsedOnSide: %s\n", side);
		te.toggleConnectionGlobal(side);
	}

	public void chiselUsedOnCentre(ShapeTE te, EntityPlayer player) {
		if (te.secondaryBlockState != null) {
			ItemStack stack = newSecondaryMaterialStack(te.secondaryBlockState);
			if (stack != null) {
				if (!Utils.playerIsInCreativeMode(player))
					Block.spawnAsEntity(te.getWorld(), te.getPos(), stack);
				te.setSecondaryMaterial(null);
			}
		}
	}
	
	protected ItemStack newSecondaryMaterialStack(IBlockState state) {
		if (acceptsCladding())
			return ArchitectureCraft.itemCladding.newStack(state, 1);
		else
			return null;
	}

	public void onHammerUse(ShapeTE te, EntityPlayer player, EnumFacing face, Vector3 hit) {
		//System.out.printf("ShapeKind.onHammerUse\n");
		if (player.isSneaking())
			te.setSide((te.side + 1) % 6);
		else
			te.setTurn((te.turn + 1) % 4);
		te.markChanged();
	}
	
	public EnumFacing zoneHit(EnumFacing face, Vector3 hit) {
		double r = 0.5 - sideZoneSize();
		//System.out.printf("ShapeKind.zoneHit: hit = (%.3f,%.3f,%.3f) r = %.3f\n",
		//	hit.x, hit.y, hit.z, r);
		if (hit.x <= -r && face != WEST) return WEST;
		if (hit.x >=  r && face != EAST) return EAST;
		if (hit.y <= -r && face != DOWN) return DOWN;
		if (hit.y >=  r && face != UP) return UP;
		if (hit.z <= -r && face != NORTH) return NORTH;
		if (hit.z >=  r && face != SOUTH) return SOUTH;
		return null;
	}
	
	public boolean acceptsCladding() {
		return false;
	}
	
	public boolean isValidSecondaryMaterial(IBlockState state) {
		return false;
	}
	
	public boolean secondaryDefaultsToBase() {
		return false;
	}

	public void addCollisionBoxesToList(ShapeTE te, IBlockAccess world, BlockPos pos, IBlockState state,
		Entity entity, Trans3 t, List list)
	{
		int mask = te.shape.occlusionMask;
		int param = mask & 0xff;
		double r;
		switch (mask & 0xff00) {
			case 0x000: // 2x2x2 cubelet bitmap
				for (int i = 0; i < 8; i++)
					if ((mask & (1 << i)) != 0) {
						Vector3 p = new Vector3(
							(i & 1) != 0 ? 0.5 : -0.5,
							(i & 4) != 0 ? 0.5 : -0.5,
							(i & 2) != 0 ? 0.5 : -0.5);
						addBox(Vector3.zero, p, t, list);
					}
					break;
			case 0x100: // Square, full size in Y
				r = param / 16.0;
				addBox(new Vector3(-r, -0.5, -r), new Vector3(r, 0.5, r), t, list);
				break;
			case 0x200: // Slab, full size in X and Y
				r = param / 32.0;
				addBox(new Vector3(-0.5, -0.5, -r), new Vector3(0.5, 0.5, r), t, list);
				break;
			default: // Full cube
				addBox(new Vector3(-0.5, -0.5, -0.5), new Vector3(0.5, 0.5, 0.5), t, list);
		}
	}

	protected void addBox(Vector3 p0, Vector3 p1, Trans3 t, List list) {
		addBox(t.p(p0), t.p(p1), list);
	}

	protected void addBox(Vector3 p0, Vector3 p1, List list) {
		AxisAlignedBB box = Trans3.boxEnclosing(p0, p1);
		list.add(box);
	}
	
	//------------------------------------------------------------------------------
	
	public static Roof Roof = new Roof();

	public static class Roof extends ShapeKind {
	
		@Override
		public boolean acceptsCladding() {
			return true;
		}
		
		@Override
		public boolean secondaryDefaultsToBase() {
			return true;
		}
	
		public void renderShape(ShapeTE te,
			ITexture[] textures, IRenderTarget target, Trans3 t,
			boolean renderBase, boolean renderSecondary)
		{
			new RenderRoof(te, textures, t, target, renderBase, renderSecondary).render();
		}
	
		@Override
		public boolean orientOnPlacement(EntityPlayer player, ShapeTE te, ShapeTE nte, EnumFacing face,
			Vector3 hit)
		{
//			if (!te.getWorld().isRemote)
//				System.out.printf("Roof.orientOnPlacement\n");
			if (!player.isSneaking() && nte != null && nte.shape.kind instanceof Roof) {
				EnumFacing nlf = nte.localFace(face);
				Profile np = profileForLocalFace(nte.shape, nlf);
				Profile p = opposite(np);
				EnumFacing lf = localFaceForProfile(te.shape, p);
				if (lf != null) {
					int turn = BaseUtils.turnToFace(lf, face.getOpposite());
//					if (!te.getWorld().isRemote) {
//						System.out.printf(
//							"Roof.orientOnPlacement: Aligning profile %s on local side %s of neighbour " +
//							"with profile %s on local side %s\n", np, nlf, p, lf);
//						System.out.printf("Roof.orientOnPlacement: Turning local side %s to face global direction %s\n",
//							lf, face.getOpposite());
//						System.out.printf("Roof.orientOnPlacement: side %s turn %s\n", nte.side, turn);
//					}
					te.setSide(nte.side);
					te.setTurn(turn);
					return true;
				}
			}
			return false;
		}
		
		protected enum Profile {None, Left, Right, Ridge, Valley};
		
		protected Profile opposite(Profile p) {
			switch (p) {
				case Left: return Profile.Right;
				case Right: return Profile.Left;
			}
			return p;
		}
		
		protected Profile profileForLocalFace(Shape shape, EnumFacing face) {
			switch (shape) {
				case RoofTile:
				case RoofOverhang:
					switch (face) {
						case EAST: return Profile.Left;
						case WEST: return Profile.Right;
					}
					break;
				case RoofOuterCorner:
				case RoofOverhangOuterCorner:
					switch (face) {
						case SOUTH: return Profile.Left;
						case WEST: return Profile.Right;
					}
					break;
				case RoofInnerCorner:
				case RoofOverhangInnerCorner:
					switch (face) {
						case EAST: return Profile.Left;
						case NORTH: return Profile.Right;
					}
					break;
				case RoofRidge:
				case RoofSmartRidge:
				case RoofOverhangRidge:
					return Profile.Ridge;
				case RoofValley:
				case RoofSmartValley:
				case RoofOverhangValley:
					return Profile.Valley;
			}
			return Profile.None;
		}
		
		protected EnumFacing localFaceForProfile(Shape shape, Profile p) {
			switch (shape) {
				case RoofTile:
				case RoofOverhang:
					switch (p) {
						case Left: return EAST;
						case Right: return WEST;
					}
					break;
				case RoofOuterCorner:
				case RoofOverhangOuterCorner:
					switch (p) {
						case Left: return SOUTH;
						case Right: return WEST;
					}
					break;
				case RoofInnerCorner:
				case RoofOverhangInnerCorner:
					switch (p) {
						case Left: return EAST;
						case Right: return NORTH;
					}
					break;
				case RoofRidge:
				case RoofOverhangRidge:
					switch (p) {
						case Ridge:
							return EAST;
					}
					break;
				case RoofValley:
				case RoofOverhangValley:
					switch (p) {
						case Valley:
							return EAST;
					}
					break;
			}
			return null;
		}

	}
	
	//------------------------------------------------------------------------------
	
	public static Model Model(String name) {
		return new Model(name, null);
	}
	
	public static Model Model(String name, Object[] profiles) {
		return new Model(name, profiles);
	}
	
	public static class Model extends ShapeKind {
	
		protected String modelName;
		
		public Model(String name, Object[] profiles) {
			this.modelName = "shape/" + name + ".smeg";
			this.profiles = profiles;
		}
	
		@Override
		public boolean secondaryDefaultsToBase() {
			return true;
		}

		public void renderShape(ShapeTE te,
			ITexture[] textures, IRenderTarget target, Trans3 t,
			boolean renderBase, boolean renderSecondary)
		{
			if (renderBase) {
				IModel model = getModel();
				model.render(t, target, textures);
			}
		}
		
		protected IModel getModel() {
			return ArchitectureCraft.mod.client.getModel(modelName);
		}
		
		@Override
		public boolean acceptsCladding() {
			BaseModel model = (BaseModel)getModel();
			for (BaseModel.Face face : model.faces)
				if (face.texture == 2)
					return true;
			return false;
		}
	
	}
	
	//------------------------------------------------------------------------------
	
	public static abstract class Window extends ShapeKind {
	
		public enum FrameKind {None, Plain, Corner};
	
		public EnumFacing[] frameSides;
		public boolean[] frameAlways;
		public FrameKind[] frameKinds;
		public EnumFacing[] frameOrientations;
		public Trans3[] frameTrans;
		
		@Override
		public boolean orientOnPlacement(EntityPlayer player, ShapeTE te, ShapeTE nte, EnumFacing otherFace,
			Vector3 hit)
		{
			int turn = -1;
			// If click is on side of a non-window block, orient perpendicular to it
			if (!player.isSneaking() && (nte == null || !(nte.shape.kind instanceof ShapeKind.Window))) {
				switch (otherFace) {
					case EAST:
					case WEST:
						turn = 0;
						break;
					case NORTH:
					case SOUTH:
						turn = 1;
						break;
				}
			}
			if (turn >= 0) {
				te.setSide(0);
				te.setTurn(turn);
				return true;
			}
			else
				return false;
		}

		public FrameKind frameKindForLocalSide(EnumFacing side) {
			return frameKinds[side.ordinal()];
		}
		
		public EnumFacing frameOrientationForLocalSide(EnumFacing side) {
			return frameOrientations[side.ordinal()];
		}

		@Override
		public boolean canPlaceUpsideDown() {
			return false;
		}

		@Override
		public double sideZoneSize() {
			return 1/8d; // 3/32d;
		}
		
		@Override
		public boolean highlightZones() {
			return true;
		}

		public void renderShape(ShapeTE te,
			ITexture[] textures, IRenderTarget target, Trans3 t,
			boolean renderBase, boolean renderSecondary)
		{
			new RenderWindow(te, textures, t, target, renderBase, renderSecondary).render();
		}
		
//		@Override
//		public void chiselUsedOnCentre(ShapeTE te, EntityPlayer player) {
//			if (te.secondaryBlockState != null) {
//				ItemStack stack = BaseUtils.blockStackWithState(te.secondaryBlockState, 1);
//				dropSecondaryMaterial(te, player, stack);
//			}
//		}

		@Override
		protected ItemStack newSecondaryMaterialStack(IBlockState state) {
			return BaseUtils.blockStackWithState(state, 1);
		}
		
		@Override
		public boolean isValidSecondaryMaterial(IBlockState state) {
			Block block = state.getBlock();
			return block == Blocks.glass_pane || block == Blocks.stained_glass_pane;
		}
		
		@Override	
		public void addCollisionBoxesToList(ShapeTE te, IBlockAccess world, BlockPos pos, IBlockState state,
			Entity entity, Trans3 t, List list)
		{
			final double r = 1/8d, s = 3/32d;
			double[] e = new double[4];
			addCentreBoxesToList(r, s, t, list);
			for (int i = 0; i <= 3; i++) {
				boolean frame = frameAlways[i] || !isConnectedGlobal(te, t.t(frameSides[i]));
				if (entity == null || frame) {
					Trans3 ts = t.t(frameTrans[i]);
					addFrameBoxesToList(i, r, s, ts, list);
				}
				e[i] = frame ? 0.5 - r : 0.5;
			}
			if (te.secondaryBlockState != null)
				addGlassBoxesToList(r, s, 1/32d, e, t, list);
		}
		
		protected void addCentreBoxesToList(double r, double s, Trans3 t, List list) {
		}

		protected void addFrameBoxesToList(int i, double r, double s, Trans3 ts, List list) {
			addBox(ts.p(-0.5, -0.5, -s), ts.p(0.5, -0.5 + r, s), list);
		}
		
		protected void addGlassBoxesToList(double r, double s, double w, double e[], Trans3 t, List list) {
			addBox(t.p(-e[3], -e[0], -w), t.p(e[1], e[2], w), list);
		}
		
		protected boolean isConnectedGlobal(ShapeTE te, EnumFacing globalDir) {
			return getConnectedWindowGlobal(te, globalDir) != null;
		}

		public ShapeTE getConnectedWindowGlobal(ShapeTE te, EnumFacing globalDir) {
			EnumFacing thisLocalDir = te.localFace(globalDir);
			FrameKind thisFrameKind = frameKindForLocalSide(thisLocalDir);
			if (thisFrameKind != FrameKind.None) {
				EnumFacing thisOrient = frameOrientationForLocalSide(thisLocalDir);
				ShapeTE nte = te.getConnectedNeighbourGlobal(globalDir);
				if (nte != null && nte.shape.kind instanceof Window) {
					Window otherKind = (Window)nte.shape.kind;
					EnumFacing otherLocalDir = nte.localFace(globalDir.getOpposite());
					FrameKind otherFrameKind = otherKind.frameKindForLocalSide(otherLocalDir);
					if (otherFrameKind != FrameKind.None) {
						EnumFacing otherOrient = otherKind.frameOrientationForLocalSide(otherLocalDir);
						if (framesMatch(thisFrameKind, otherFrameKind,
							te.globalFace(thisOrient), nte.globalFace(otherOrient)))
								return nte;
					}
				}
			}
			return null;
		}
		
		protected boolean framesMatch(FrameKind kind1, FrameKind kind2,
			EnumFacing orient1, EnumFacing orient2)
		{
			if (kind1 == kind2) {
				switch (kind1) {
					case Plain:
						return orient1.getAxis() == orient2.getAxis();
					default:
						return orient1 == orient2;
				}
			}
			return false;
		}

//		protected EnumFacing getFrameOrientationGlobal(ShapeTE te, EnumFacing globalDir) {
//			Trans3 t = te.localToGlobalRotation();
//			EnumFacing localDir = t.it(globalDir);
//			return frameOrientations[localDir.ordinal()];
//		}
	
	}

	//------------------------------------------------------------------------------
	
	public static Cladding Cladding = new Cladding();

	public static class Cladding extends ShapeKind {
	
		public void renderShape(ShapeTE te,
			ITexture[] textures, IRenderTarget target, Trans3 t,
			boolean renderBase, boolean renderSecondary)
		{}
		
		public ItemStack newStack(Shape shape, Block materialBlock, int materialMeta, int stackSize) {
			return ArchitectureCraft.itemCladding.newStack(materialBlock, materialMeta, stackSize);
		}
	
	}
		
	//------------------------------------------------------------------------------
}
